import re
occurances=re.finditer("ab", "abcabcaabbc")
[(m.start(),m.end()) for m in occurances]
[m.span() for m in occurances]
[m.start() for m in occurances]

def naive_matches(pattern, text):
    """ naive search for exact matches of pattern inside text
    returns the starting points of all pattern occurances """
    assert len(pattern) <= len(text)  # lengths are compatible
    n = len(text)
    m = len(pattern)
    matches = []   # will hold all locations of matches start indices
    for i in range(n-m+1):  # do not forget the +1
        if pattern == text[i:i+m]:
            matches = matches+[i]
    return matches


def fingerprint(string, basis=2**16, r=2**32-3):
    """ used to computes karp-rabin fingerprint of the pattern
    employs Horner method (modulo r) """
    partial_sum = 0
    for x in string:
        partial_sum = (partial_sum*basis + ord(x)) % r
    return partial_sum

def text_fingerprint(string, length, basis=2**16, r=2**32-3):
    """ used to computes karp-rabin fingerprint of the text """
    f = []
    b_power = pow(basis,length-1,r)
    list.append(f, fingerprint(string[0:length], basis, r))
    # f[0] equals first text fingerprint 
    for s in range(1,len(string)-length+1):
        new_fingerprint = ((f[s-1] - ord(string[s-1])*b_power)*basis
                         +ord(string[s+length-1])) % r
            # compute f[s], based on f[s-1]
        list.append(f,new_fingerprint)# append f[s] to existing f       
    return f

def find_matches_KR(pattern, text, basis=2**16, r=2**32-3):
    if len(pattern) > len(text):
        return []
    p = fingerprint(pattern,basis,r)
    f = text_fingerprint(text,len(pattern),basis,r)
    matches = [s for s, f_s in enumerate(f) if f_s == p]
    # list comprehension 
    return matches

def find_matches(pattern,text,basis=2**16):
    """ find all occurances of pattern in text
        using efficient arithmetization of text """
    assert(len(pattern) <= len(text))
    p = arithmetize(pattern,basis)
    t = arithmetize_text2(text,len(pattern),basis)
    matches = []
    for s in range(len(t)):
        if p==t[s]:
            matches = matches + [s]
    return matches

def arithmetize_text2(text, m, basis=2**16):
    """  efficiently computes arithmization of all m long 
         substrings of text, using powers of basis """
    b_power = basis**(m-1)
    t = [arithmetize(text[0:m], basis)] # t[0] equals first word arithmetization
    # will store list of numbers representing consecutive m long chuncks of text
    for s in range(1,len(text)-m+1):
        new_number = ((t[s-1] - ord(text[s-1])*b_power)*basis +
                      ord(text[s+m-1]))
        t = t + [new_number]  # append new_number to existing t
    return t
 


def arithmetize(string, basis=2**16):
    """ convert substring to number using basis powers
        employs Horner rule """
    partial_sum = 0
    for ch in string:
        partial_sum =(partial_sum*basis+ord(ch)) # Horner
    return partial_sum
    

        
def arithmetize_text(text, m, basis=2**16):
    """  computes arithmization of all m long subwords
         of text, using basis powers """
    t = []
    # will store list of numbers representing consecutive chuncks of string
    for s in range(0,len(text)-m+1):
        t = t + [arithmetize(text[s:s+m],basis)]
        # append the next number to existing t
    return t
        


        
